import { logger } from "../log.js";
const net = require("net")
  , tls = require("tls")
  , EventEmitter = require("events").EventEmitter
  , util = require("util");

class irc {
  constructor(options) {
    EventEmitter.call(this);
    this.options = options || {};
    this.options.channels = this.options.channels || [];
    this.options.host = this.options.host || "127.0.0.1";
    this.options.port = this.options.port || 6667;
    this.options.ssl = this.options.ssl || false;
    this.options.selfSigned = this.options.selfSigned || false;
    this.options.sasl = this.options.sasl || false;
    this.network = this.options.network || "test";
    this.nickname = this.options.nickname || "test";
    this.username = this.options.username || "test";
    this.realname = this.options.realname || "test";
    this.channels = this.options.channels || [];
    this._recachetime = 60 * 5; // 10 minutes
    this.server = {
      motd: "",
      me: {},
      channel: [],
      user: new Map()
    };
    this.socket = (this.options.ssl ? tls : net).connect({
      host: this.options.host,
      port: this.options.port,
      rejectUnauthorized: !this.options.selfSigned
    }, () => {
      this.send(`NICK ${this.nickname}`);
      this.send(`USER ${this.username} 0 * : ${this.realname}`);
      if(this.options.sasl)
        this.send("CAP LS");
    });
    this.socket.setEncoding("utf-8");
    this.socket.on("data", msg => {
      msg = msg.split(/\r?\n|\r/);
      for(let tmp in msg)
        if(msg[tmp].length > 0)
          this.cmd(this.parse(msg[tmp]));
    });
  }
  send(data) {
    this.socket.write(`${data}\n`);
    logger.debug(`out: ${data}`);
  }
  parse(data, [a, ...b] = data.split(/ +:/)) {
    let tmp = a.split(" ").concat(b);
    logger.debug(`in: ${[...tmp]}`);
    return data.charAt(0) === ":" ? {
      prefix: tmp.shift(),
      command: tmp.shift(),
      params: tmp
    } : {
      prefix: null,
      command: tmp.shift(),
      params: tmp
    };
  }
  cmd(msg) {
    let tmpuser = {};
    let tmpchan = {};
    let chans = [];
    let prefix = "";
    switch (msg.command) { // auslagern!
    case "307": // Rizon Account
      tmpuser = {};
      if (this.server.user.hasi(msg.params[1]))
        tmpuser = this.server.user.geti(msg.params[1]);
      tmpuser.account = msg.params[1];
      tmpuser.registered = true;
      this.server.user.set(msg.params[1], tmpuser);
      break;
    case "311": // first
      tmpuser = {};
      if (this.server.user.hasi(msg.params[1]))
        tmpuser = this.server.user.geti(msg.params[1]);
      tmpuser.nickname = msg.params[1];
      tmpuser.username = msg.params[2];
      tmpuser.hostname = msg.params[3];
      tmpuser.realname = msg.params[5];
      this.server.user.set(msg.params[1], tmpuser);
      break;
    case "313": // Oper
      tmpuser = {};
      if (this.server.user.hasi(msg.params[1]))
        tmpuser = this.server.user.geti(msg.params[1]);
      tmpuser.oper = true;
      this.server.user.set(msg.params[1], tmpuser);
      break;
    case "318": // last (check Data)
      tmpuser = {};
      if (this.server.user.hasi(msg.params[1]))
        tmpuser = this.server.user.geti(msg.params[1]);
      tmpuser = {
        nickname: tmpuser.nickname || false,
        username: tmpuser.username || false,
        hostname: tmpuser.hostname || false,
        realname: tmpuser.realname || false,
        account: tmpuser.account || false,
        registered: tmpuser.registered || false,
        oper: tmpuser.oper || false,
        channels: tmpuser.channels || [],
        cached: ~~(Date.now() / 1000)
      };
      this.server.user.set(msg.params[1], tmpuser);
      break;
    case "319": // Chanlist Todo
      tmpchan = new Map();
      tmpuser = {};
      if (this.server.user.hasi(msg.params[1])) {
        tmpuser = this.server.user.geti(msg.params[1]);
        if (tmpuser.channels)
          tmpchan = tmpuser.channels;
      }
      chans = msg.params[2].trim().split(" ");
      for (let chan in chans) {
        chan = chans[chan].split("#");
        tmpchan.set(`#${chan[1]}`, chan[0]);
      }
      tmpuser.channels = tmpchan;
      this.server.user.set(msg.params[1], tmpuser);
      break;
    case "330": // Quarknet
      tmpuser = {};
      if (this.server.user.hasi(msg.params[1]))
        tmpuser = this.server.user.geti(msg.params[1]);
      tmpuser.account = msg.params[2];
      tmpuser.registered = true;
      this.server.user.set(msg.params[1], tmpuser);
      break;
    case "001": // welcome
      this.server.me = msg.params[0];
      this.join(this.options.channels);
      this.emit("data", ["connected", msg.params[1]]);
      break;
    case "352": // who_entry
      if (!this.server.channel[msg.params[1]])
        this.server.channel[msg.params[1]] = new Map();
      this.server.channel[msg.params[1]].set(msg.params[5], { // chan
        nick: msg.params[5],
        username: msg.params[2],
        hostname: msg.params[3]
      });
      this.whois(msg.params[5]);
      break;
    case "315": // who_end
      //console.log(this.server.channel);
      break;
    case "372": // motd
      this.server.motd += `${msg.params[1]}\n`;
      break;
    case "375": // motd_start
      this.server.motd = `${msg.params[1]}\n`;
      break;
    case "376": // motd_end
      this.server.motd += `${msg.params[1]}\n`;
      this.emit("data", ["motd", this.server.motd]);
      break;
    case "464": // Password required
      if (this.options.password.length > 0 && !this.options.sasl)
        this.send(`PASS ${this.options.password}`); // pwd
      break;
    case "PING":
      this.send(`PONG ${msg.params.join``}`);
      break;
    case "JOIN": // wip
      this.send(`WHO ${msg.params[0]}`);
      break;
    case "PART": // wip
      delete this.server.user[msg.params[0]];
      break;
    case "PRIVMSG":
      this.emit("data", msg.params[1] === "\u0001VERSION\u0001" ? ["ctcp:version", this.reply(msg)] : ["message", this.reply(msg)]);
      break;
    case "NOTICE":
      this.emit("data", ["notice", msg.params[1]]);
      break;
    case "NICK": // nickchange
      prefix = parsePrefix(msg.prefix);
      if(this.server.user.hasi(prefix.nick))
        this.server.user.deli(prefix.nick);
      this.whois(msg.params[0], true); // force
      break;
    case "CAP":
      switch (msg.params[1]) {
      case "LS":
        this.send(`CAP REQ :${msg.params[2]}`);
        break;
      case "ACK": // success
        this.send("AUTHENTICATE PLAIN");
        break;
      }
      break;
    case "AUTHENTICATE":
      if (msg.params[0].match(/\+/))
        this.send(`AUTHENTICATE ${new Buffer(this.username + "\u0000" + this.username + "\u0000" + this.options.password).toString("base64")}`);
      break;
    case "900":
      this.send("CAP END");
      break;
    }
  }
  reply(tmp) {
    return {
      type: "irc",
      network: this.network,
      channel: tmp.params[0],
      user: parsePrefix(tmp.prefix),
      message: tmp.params[1],
      time: ~~(Date.now() / 1000),
      raw: tmp,
      reply: msg => this.send(`PRIVMSG ${tmp.params[0]} :${msg}`),
      replyAction: msg => this.send(`PRIVMSG ${tmp.params[0]} :\u0001ACTION ${msg}\u0001`),
      replyNotice: msg => this.send(`NOTICE ${tmp.params[0]} :${msg}`),
      _chan: this.server.channel[tmp.params[0]],
      _user: this.server.user,
      join: chan => this.join(chan),
      part: (chan, msg) => this.part(chan, msg),
      whois: user => this.whois(user),
      write: msg => this.send(msg)
    };
  }
  join(channel) {
    this.send(`JOIN ${(typeof channel === "object") ? channel.join(",") : channel}`);
  }
  part(channel, msg=false) {
    this.send(`PART ${(typeof channel === "object") ? channel.join(",") : channel}${msg ? " " + msg : " part"}`);
  }
  whois(user, force = false) {
    user = user.toLowerCase();
    if(this.server.user.hasi(user) && !force)
      if(this.server.user.geti(user).cached >= ~~(Date.now() / 1000) - this._recachetime)
        return;
    this.send(`WHOIS ${user}`);
  }
}

module.exports = irc;
util.inherits(irc, EventEmitter);

function parsePrefix(prefix) {
  let usertmp = prefix.split("!");
  let hosttmp = usertmp[1].split("@");
  return {
    nick: usertmp[0],
    username: hosttmp[0],
    hostname: hosttmp[1]
  };
}

Map.prototype.hasi = function(val) {
  for (let [key] of this)
    if(key.toLowerCase() === val.toLowerCase())
      return true;
  return false;
};
Map.prototype.geti = function(val) {
  for (let [key, value] of this)
    if(key.toLowerCase() === val.toLowerCase())
      return value;
  return false;
};
Map.prototype.deli = function(val) {
  for (let [key] of this)
    if(key.toLowerCase() === val.toLowerCase())
      this.delete(key);
};